"""
Load and save variable data, persistent options to the data file.

modes:
    play                - track | dir | playlist
    continuous          - stop playing if not set
    random              - track | dir | off
    random_norepeat    - create a queue and randomly play it until 
                          exhausted

options:
    threshold           - see README
    check_for_updates   - check online for new versions
    random_slope        - 1-5; set higher to play high-scored
                          tracks more often in random mode
    (if you're using random mode and you play song A, and you don't want to hear
    song A again right after it, set this to 1, or a higher number)

scores              - dictionary of scores {fname: score, ..}
last_track          - last track played (to restart on it)
"""

from avkutil import debug
import shelve
import conf

_vars = """play interval continuous last_track playlists threshold random
random_norepeat scores randslope last_check_nv queue lastplaylist""".split()

class Data:
    """Borg design pattern (all instances share the same state). See
    ActiveState python cookbook for this recipe.
    """

    _shared = {}    # shared state 


    def __init__(self):
        self.__dict__ = self._shared


    def load(self):
        # set defaults:
        self.last_check_nv      = None  # last time checked for new version
        self.scores             = None
        self.play               = "playlist"
        self.continuous         = 1
        self.random             = "off"
        self.random_norepeat    = 0
        self.threshold          = 0    
        self.randslope          = 2       
        self.playlists          = {}
        self.scores             = {}
        self.lastplaylist       = None
        self.last_track         = None
        self.randlist           = []
        self.queue              = []
        self.interval           = 0     # pause between tracks

        self.sh = shelve.open(conf.data)

        for opt in self.sh.keys():
            self.__dict__[opt] = self.sh[opt]


    def save(self):
        for opt in _vars:
            self.sh[opt] = self.__dict__[opt]
        self.sh.close()


data = Data()
