''' Playlist class
'''

import os
import copy
import time
import curses
import sys
import threading
import string
import avkutil
import conf
import main
from colors import colors
import ui
from status import stat
from data import data
from track import Track
from album import Album
from player import player
from scores import scores
from log import log
from cycurses import cycurses
from debug import debug
import flags

PlaylistLoadError = """
Error reading playlist file

You could be trying to load a non-cymbaline playlist or one from an older
version of cymbaline.
"""

class Loadtimes(threading.Thread):
    """Load time info for tracks in current album."""

    def __init__(self, albums, start, end):
        threading.Thread.__init__(self)
        self.albums = albums
# tstart because self.start() starts the thread (doh)
        self.tstart = start
        self.end = end

    def run(self):
        redraw = 0
        time.sleep(2)
        if data.random == 'queue':
            n = 1
            for tmp, ai, i in data.queue:
#                if i in (0,1,2):
                time.sleep(0.1)
                if player.do_quit:
                    return
                try:
                    updated = self.albums[ai][i].info()
                except IndexError:
# this means one track was deleted..
                    pass
                if ui.tracks_window and updated:
                    flags.ScreenUpdated = 1
                n+=1
        else:
            for track in pl.tracks[self.tstart:self.end]:
                try:
                    # throws exception when quitting, time = None?
                    time.sleep(0.1)
                except:
                    return None
                updated = track.info()
                if updated and ui.tracks_window: 
                    flags.ScreenUpdated = 1
            #ui.fullredraw()


class Playlist:
    """Cymbaline playlist.

    Usage: self.update(0,0) sets current album and track to 0. self.update(3,5)
    sets current album 3 and current track 5. Then you'd have to call
    player.play() to set it in motion.

    There are two "backbone" data structures - albums and tracks, the latter is
    a list of tracks in current album. Two corresponding indexes are albumindex
    and index.

    On top of that, there are convenience shortcuts like self.album (current
    album), self.tracks (tracks of the current album) and self.track (current
    track) - these are all updated by self.update() from albums, tracks, index
    and albumindex.
    """

    _shared = {}       # borg (singleton) design pattern

    def __init__(self):
        self.__dict__ = self._shared


    def init(self, dirs):
        """Initialize playlist.

        index               - track index in current album
        albumindex          - album index in list of albums
        """

        self.last_i, self.last_ai = None, None
        self.albumindex, self.index = 0, 0
        self.do_play = 1
        self.dirs = dirs
        self.last_search = None
        self.albums      = []
        self.full_list   = []
        self.tracksnum   = 0
        okay = self.loaddirs(dirs)
        if not okay:
            return 0
        self.update(0,0)
        time.sleep(0.1)
        return 1

    
    def seek(self):
        '''Seek to given value in min:seconds.'''
        ui.doredraw = 0
        ui.cl()
        try:
            exact = ui.ask('Seek to (min:sec or just sec): ')
            if exact.find(':') != -1:
                flds = exact.split(':')
                mins = int(flds[0])*60
                secs = int(flds[1])
                exact = mins + secs
            else:
                exact = int(exact)
        except:
            stat.error('Invalid input.')
            return
        player.seek(exact=exact)
    
    def interval(self):
        '''Set interval (pause) between songs to a new value.'''
        ui.doredraw = 0
        ui.cl()
        try:
            data.interval = int(ui.ask('Enter new interval value: '))
        except ValueError:
            stat.error('Invalid input.')
        ui.doredraw = 1


    def list_albums(self):

        i = self.albumindex
        h = ui.h()-3
        l = len(self.albums)
        start = 0
        end = 9999
        if l <= h:
            pass
        else:
            # trim track list to fit in screen, with current track visible
            if i < h/2:
                start = 0
                end = h
            elif l-i < h/2:
                start = l - h
                end = l
            else:
                start = i - h/2
                end = i + h/2 + 1

        ui.tracks_window = 0
        w = cycurses.win
        w.clear()
        # ui.do_clear = 1
        n = 1
        st = start
        for album in self.albums[start:end]:
            st += 1
            w.addnstr(n, 0, "%2d) " % st, 10)
            if album == self.album and curses.has_colors():
                w.addnstr(n, 5, repr(album.name)[1:-1], ui.w()-6, colors.cur_dir)
            else:
                w.addnstr(n, 5, repr(album.name)[1:-1], ui.w()-6)
            n+=1
        w.refresh()
        ui.do_clear = 1

    
    def list_playlists(self):
        """List playlist files in ~/.cymbaline/playlists/ directory."""
        ui.tracks_window = 0
        ui.do_clear = 1
        playlists = data.playlists.keys()
        show = string.join(playlists, '\n')
        show = 'Playlists:\n\n' + show
        ui.fsdisplay(show)
        ui.do_clear = 1


    def show_path(self):
        stat.normal(self.track.path)


    def slp(self):
        '''Sleep for debugging purposes.'''
        if conf.debug:
            ui.doredraw=0
#            time.sleep(2) #dbgsleep
            ui.doredraw=1


    def ls(self, one_screen=1):
        if data.queue:
            current = data.queue[0]
        tmp = data.queue[:4]
#        debug("LS: data.queue[:4]: %s, %s, %s, %s, self.track.name: %s, current: %s\n" % (tmp[0][1:], tmp[1][1:], 
#                    tmp[2][1:], tmp[3][1:], self.track.name, current[1:]))
        w = cycurses.win
        lst = []
        n = 1
        if data.random == 'queue':
            tracks = []
            sum = 0
            for tmp in data.queue:
                # track inst., dirindex
                tracks.append((tmp[0], tmp[1]))     
                sum += tmp[0].score() or conf.def_score
            avg = sum / float(len(data.queue))

            txt = 'Random Queue (avg score: %d)' % avg
            w.addnstr(0, 0, txt.center(ui.w()), ui.w()-1)
        else:
            tracks = self.tracks
            sum = 0
            for tr in tracks:
                sum += tr.score() or conf.def_score
            avg = sum / float(len(tracks))
            txt = '%s (avg score: %d)' % (self.album.name.strip(), avg)
            if curses.has_colors():
                w.addnstr(0, 0, txt.center(ui.w()), ui.w()-1, colors.dir_name)
            else:
                w.addnstr(0, 0, txt.center(ui.w()), ui.w()-1)


        h = ui.h()-3
        if data.random == 'queue':
            # in queue current track is always first, and self.index points
            # to whatever number it is in its album
            i = 0
            l = len(data.queue)
        else:
            l = len(self.album)
            i = self.index

        start = 0

        blank = None
        if l <= h:
# blank these lines
            start = 0
            end = l
            blank = (h+2) - l
        else:
            # trim track list to fit in screen, with current track visible
            newlst = []
            if i < h/2:
                start = 0
                end = h
            elif l-i < h/2:
                start = l - h
                end = l
            else:
                start = i - h/2
                end = i + h/2 + 1
            newlst = tracks[start:end]
            tracks = newlst

# these are for Loadtimes object
        self.TracksStart = start
        self.TracksEnd = end
        Loadtimes(self.albums, self.TracksStart, self.TracksEnd).start()
        current_hl = 0
        for i in range(len(tracks)):
            track = tracks[i]
            if data.random == "queue":
                track = tracks[i][0]

            if track.fname == self.track.fname and not current_hl:
                w.addnstr(i+1, 0, '> ', 2)

                current_hl = 1
                colr = colors.cur_track
                br = colors.current_bracket
            else:
                w.addnstr(i+1, 0, '  ', 2)
                colr = colors.track_name
                br = colors.default
            track.id3name()
# I already do that in track.py
#            if data.random == 'queue' and track.artist_title:
#                name = track.artist_title
            name = track.name
            if data.random == "queue":
                di = tracks[i][1]
                album = pl.albums[di]
                if " - " in name:
                    pass
                elif track.artist:
                    x = track.artist
                    x = string.capwords(x)
                    name = "%s - %s" % (x, name)
                elif album.artist:
                    x = album.artist
                    x = string.capwords(x)
                    name = "%s - %s" % (x, name)
                    debug ("album.artist:" + album.artist + "\n")
                else:
                    x = album.justName.strip()
                    x = string.capwords(x)
                    name = "%s - %s" % (x, name)

            name = name[:ui.w()-20].ljust(ui.w()-20)
            sc, scolor = scores.color_score(track) 
            if track.total == None:
                total = ' '*5
            else:
                tot = avkutil.ftime(track.total)
                if len(tot)>5:
                    tot = "99:99"
                total = '%5s' % tot
            start += 1

            if curses.has_colors():
                w.addstr(i+1, 2, '%2d) ' % (start), colors.default)
                #debug("i+1: %s, 6, name: %s, colr: %s\n" % (i+1, name, colr))
                w.addstr(i+1, 6, name, colr)
#                except:
#                    debug("error in: i+l %d, 6, name %s, colr %s\n" % (i+l, name, colr))
#                    main.quit()
#                    sys.exit()
                w.addstr(i+1, ui.w()-20+6, sc + ' ' , scolor)
                w.addstr(i+1, ui.w()-20+6+6, '[', br)
                w.addstr(i+1, ui.w()-20+6+6+1, total, colors.time)
                w.addstr(i+1, ui.w()-20+6+6+1+5, ']', br)
            else:
                w.addstr(i+1, 2, '%2d) ' % (i+1))
                w.addstr(i+1, 6, name)
                w.addstr(i+1, ui.w()-20+6, sc)
                w.addstr(i+1, ui.w()-20+6+6, '[')
                w.addstr(i+1, ui.w()-20+6+6+1, total)
                w.addstr(i+1, ui.w()-20+6+6+1+5, ']')

# now we need to blank out the lines that are left off after tracks:
        if blank:
            last = i+2
            for i in range(blank):
                w.addstr(i + last, 0, ' '*(ui.w()-1))
        # w.refresh()


    def save(self):
        """Save playlist.

        First we'll try to guess the name from dirname, i.e. if it's
        /music/pop, a good name would be 'pop'.
        """

        tmp = self.albums[0].path.split('/')
        default = tmp[-1] or tmp[-2]
        name = ui.ask("Save playlist as[%s]: " % default)
        if not name:
            name = default
        #oldname = self.name
        self.name = name
        data.playlists[name] = self
        data.lastplaylist = name
        stat.normal("playlist saved.")


    def update(self, ai=None, i=None, quick=0):
        """Update playlist shortcuts.

        quick option is for searching function.
        Update 'shortcut' variables. See Playlist.__doc__ for details.
        """

        self.last_ai, self.last_i = self.albumindex, self.index
#        if ai != self.albumindex and data.random != 'queue' and not quick:
#            ui.clear()
        if ai == None: ai = self.albumindex
        if i == None: i = self.index
        self.index      = i
        self.albumindex = ai
        self.album      = self.albums[self.albumindex]
        self.track      = self.album[self.index]
        self.tracks     = self.album.tracks
#        if not quick:
#            Loadtimes(self.albums).start()
        if self.track.secs == None and not quick:
            # load time info
            self.track.info()


    def manual_next(self, seek=1):
        """Next track set off by user.
        
        It makes sense I guess to only decrease score if the track was skipped
        to next one (seek=1). Otherwise it may be that the user likes current
        track just fine but wants to seek to a particular track in the album.
        """

        if seek == 1:
            self.next(seek=seek, lower_score=1)
        else:
            self.next(seek=seek, lower_score=0)


    def next(self, seek=1, play=1, lower_score=0):
        """Next track."""
        import rand
        if seek == 1 and self.do_play and not lower_score:
            # track finished fully: increase score!
            scores.set(conf.score_increment, absolute=0)
        elif lower_score:
            # skipped track, decrease score
            scores.set(-conf.score_increment, absolute=0)
        stat.normal("next")
        i = self.index + seek
        if data.random != "off":
            if seek == 1:
                if data.random == "track":
                    #raw_input('\n')
                    rand.rand()
                    return
                elif data.random == "album":
                    if i >= len(self.album):
                        rand.random_album()
                        return
                    else:
                        # keep playing this album
                        pass
            if data.random == "queue":
                ql = data.queue
                if ql:
                    del ql[:seek]
                rand.queue()
                current = ql[0]
                self.update(current[1], current[2])
                if play:
                    player.play()
                tmp = data.queue[:4]
                # debug("data.queue[:4]: %s, %s, %s, %s, self.track.name: %s, current: %s\n" % (tmp[0][1:], tmp[1][1:], 
                    # tmp[2][1:], tmp[3][1:], self.track.name, current[1:]))
                return

        if i < len(self.album):
            quick = not play
            self.update(self.albumindex, i, quick=quick)
            if play:
                player.play()
            #self.slp()
        else:
            if seek == 1:
                if data.play == 'dir':
                    self.first()
                else:
                    self.next_album(play=play)
            else:
                self.last()
        #time.sleep(1.5)


    def prev(self, seek=1, play=1):
        """Previous track."""
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        i = self.index - seek
        if i >= 0:
            self.update(self.albumindex, i)
            if play:
                player.play()
        else:
            if seek == 1:
                self.prev_album(track="last", play=play)
            else:
                self.first()
        stat.normal("previous")


    def first(self):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        self.update(self.albumindex, 0)
        player.play()
        stat.normal("first song")


    def last(self):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        self.update(self.albumindex, len(self.album)-1)
        player.play()
        stat.normal("last song")


    def Nth(self, seek):
        if seek >= len(self.album):
            seek = len(self.album)
        self.update(self.albumindex, seek-1)
        player.play()
        stat.normal("song #%d" % seek)


    def next_album(self, seek=1, play=1):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        if data.random == 'album':
            import rand
            rand.random_album()
            return
        di = self.albumindex + seek
        if di >= len(self.albums):
            if seek == 1:
                di = 0
            else:
                di = len(self.albums)-1
        quick = not play
        self.update(di, 0, quick=quick)
        if play:
            player.play()
        if play:
            stat.normal("next album")


    def prev_album(self, seek=1, track="first", play=1):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        di = self.albumindex - seek
        if track == "first":
            i = 0
        else:
            i = len(self.albums[di])-1
        if di >= 0:
            quick = not play
            self.update(di, i, quick=quick)
        else:
            if seek == 1:
                self.last_album(track=track)
            else:
                self.first_album(track=track)
        if play:
            player.play()
        stat.normal("previous album")


    def first_album(self, track="first"):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        if track == "first":
            i = 0
        else:
            i = len(self.albums[0])-1
        self.update(0, i)
        player.play()
        stat.normal("first album")


    def last_album(self, track="first"):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        di = len(self.albums)-1
        if track == "first":
            i = 0
        else:
            i = len(self.albums[di])-1
        self.update(di, i)
        player.play()
        stat.normal("last album")


    def Nth_album(self, seek):
        if data.random == "queue":
            stat.error("Disabled in random queue mode.")
            return
        if seek >= len(self.albums):
            self.last_album()
        else:
            self.update(seek-1, 0)
            player.play()
        aind = self.albumindex + 1
        stat.normal("album #%d" % aind)


    def find_highscore(self, thres=None):
        """Find next track with high_score or higher."""
#        if data.random == "queue":
#            stat.error("Disabled in random queue mode.")
#            return
        if not thres:
            thres = conf.highscore
        i, di = self.index, self.albumindex
        player.mode = "stop"
        self.do_play = 0
        while 1:
            self.next(play=0)
            s = self.track.score() or conf.def_score
            if s >= thres:
                self.do_play = 1
                player.play()
                ui.fullredraw()
                break
            if (self.index, self.albumindex) == (i, di):
                stat.error("No high scored tracks found.")
                self.do_play = 1
                break


    def toggle_threshold(self):
        """Toggle threshold between 0 and conf.threshold_toggle."""
        if not data.threshold:
            self.set_threshold(conf.threshold_toggle)
        else:
            self.set_threshold(0)
        stat.normal("threshold = %d" % data.threshold)


    def set_threshold(self, thres):
        """Set threshold to thres.

        Only tracks with score >= threshold will be played. List of
        directories (dirs) will have to be updated - if threshold is
        increased, tracks with low score will be pruned; if it's
        decreased, tracks with low score will be added. Full dirs
        is stored in full_list. This happens when we change from
        threshold=0; when it's 0 we know that dirs is full - no
        songs were pruned from it.

        Therefore, if threshold increases, we step through the playlist
        and remove any songs with low score. When it decreases, we do
        the same but first we have to load up all songs from
        full_list. And if thres=>0, we just load fulldict.
        """

        name = self.track.name
        if not data.threshold:
            # 0 => 1+
            self.full_list = copy.deepcopy(self.albums)
        if thres < data.threshold:
            # high => low
            self.albums = copy.deepcopy(self.full_list)
            if not thres:
                data.threshold = thres
                self.find(pattern=name)
                # self.find plays it
                # player.play()
                return
        nl = []
        for album in self.albums:
            tmp = []
            for track in album.tracks:
                if (track.score() or conf.def_score) >= thres:
                    tmp.append(track)
            if tmp:
                album.tracks = tmp
                nl.append(album)
        if nl:
            self.albums = nl
            data.threshold = thres
            i = di = 0
            self.update(di, i, quick=1)
            if not self.find(pattern=name):
                self.first_album()
                player.play()
            stat.normal("threshold = %d " % thres )
        else:
            stat.error("Change of threshold yielded empty playlist.")


    def find(self, direction=1, pattern='', show_error=1, OnlyFind=0):
        """Find a track in playlist.
        
        set direction to 0 to search backwards; show_error will show
        error in status bar when no track is found; OnlyFind set to 1 to
        avoid playing it after it is found.
        """

        was_rqueue = 0
        if data.random == 'queue':
            '''Searching in random queue makes little sense 'cause what if it's
            not there? Huh? '''
            data.random = 'off'
            was_rqueue = 1
        # TODO: time if searching path takes much longer than name
        if not pattern:
            pattern = ui.ask("/")
        pattern = pattern.lower()
        i, di = self.index, self.albumindex
        self.do_play = 0
        while 1:
            if i+1 >= len(self.albums[di]):
                if di+1 >= len(self.albums):
                    di = 0
                else:
                    di += 1
                i = 0
            else:
                i += 1
            if (self.albums[di][i].name.lower().find(pattern) > -1 or
                        self.albums[di][i].fname.lower().find(pattern) > -1):
                self.update(di, i)
                self.do_play = 1
                if not OnlyFind:
                    player.play()
                return 1
            if (self.index, self.albumindex) == (i, di):
                self.do_play = 1
                if show_error:
                    stat.error("Not found: " + pattern)
                if was_rqueue:
                    data.random = "queue"
                return 0


    def loaddirs(self, dirlist):
        """Recursively add mp3s from dirlist to playlist."""

        # TODO: store only filename, path is stored in dir, join them
        # in Play() - to save space??
        def visit(self, dirpath, names):
            lst = []
            names.sort()
            for name in names:
                ext = name.split('.')[-1].lower()
                codec = None
                if ext in conf.mp3:
                    codec = "mp3"
                #elif ext in conf.ogg:
                #    codec = "ogg"
                #elif ext in conf.mod:
                #    codec = "mod"
                if codec:
                    path = os.path.join(dirpath, name)
                    lst.append(Track(path, codec))
            if not lst:
                return
            self.tracksnum += len(lst)
            if dirpath[0] != '/':
                dirpath = os.path.join(os.getcwd(), dirpath)
            if dirpath[-1] == '/':
                dirpath = dirpath[:-1]
            album = Album(dirpath, lst)
            self.albums.append(album)

        for directory in dirlist:
            os.path.walk(directory, visit, self)

        if not self.albums:
            ui.print_line(5, ("  Could not find any mp3 files! Exiting..", 0))
            ui.refresh()
            time.sleep(2)
            return 0
        return 1


    def html(self):
        """Write playlist to an html file."""

        ui.doredraw = 0
        ui.cl()
        filename = ui.ask('Filename to save to (.html is added automatically): ')
        ui.doredraw = 1
        stat.normal("writing playlist to %s.html" % filename)
        f = open(os.path.expanduser("~/" + filename + ".html"), 'w')
        f.write("""<HTML><HEAD><STYLE><!-- BODY {FONT-FAMILY: ARIAL,
            SANS-SERIF;} //--> </STYLE></HEAD><BODY BGCOLOR="%s">\n\n
            """ 
            % colors.htmlbgcolor)
        f.write("<CENTER><H3>My Playlist</H3>\n")
        f.write("<table> <tr> <td colspan=2> Generated by cymbaline.py on %s, threshold %d <P><BR> </td></tr> \n" % 
            (time.ctime(time.time()), data.threshold))
        f.write(" <tr> <td>Name</td> <td> Score </td> </tr>")
        for album in self.albums:
            f.write("""<tr> <td><p><BR><font color="blue">%s</font> </td> <td></td>
            </tr> \n""" % album)
            for track in album.tracks:
                f.write("""<tr> <td> <font color="green">%s</font> </td> <td
                align=right> %s </td> </tr>\n""" % (track.name, str(track.score())))
        f.write("</table> </center> </body></html>\n")
        f.close()


    def setslope(self):
        slope = ui.ask('Enter random slope from 1 to 10: ')
        try:
            sl = int(slope)
        except:
            stat.error('Invalid input.')
            return
        if not 1 <= sl <= 10:
            stat.error("Random slope must be from 1 to 10.")
        else:
            data.randslope = sl
            ui.fullredraw()
            stat.normal("slope = %d" % sl)


    def reload(self):
        """Reload current playlist from disk."""
        stat.normal("Reloading...")
        name = self.track.name
        self.tracksnum = 0
        self.albums, self.full_list = [], []
        ok = self.init(self.dirs)
        if not ok:
            stat.error("Error reloading playlist, exiting..")
            main.quit()
        self.update(0,0)
        self.find(1, name)
        stat.normal("Playlist reloaded from disk.")
        ui.fullredraw()


pl = Playlist()
